"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SocketServer = void 0;
const tslib_1 = require("tslib");
const ws_1 = require("ws");
const types_1 = require("tdf-devtools-protocol/dist/types");
const enum_1 = require("@/@types/enum");
const db_1 = require("@/db");
const client_1 = require("@/client");
const debug_targets_1 = require("@/controller/debug-targets");
const pub_sub_manager_1 = require("@/controller/pub-sub-manager");
const log_1 = require("@/utils/log");
const pub_sub_channel_1 = require("@/utils/pub-sub-channel");
const url_1 = require("@/utils/url");
const debug_target_1 = require("@/utils/debug-target");
const config_1 = require("@/config");
// 断开连接后不再发送调试指令，不会出现 id 混乱，所以 command id 可以 mock 一个
const mockCmdId = -100000;
const resumeCommands = [
    {
        id: mockCmdId,
        method: types_1.TdfCommand.TDFRuntimeResume,
        params: {},
    },
    {
        id: mockCmdId - 1,
        method: types_1.ChromeCommand.DebuggerDisable,
        params: {},
    },
    {
        id: mockCmdId - 2,
        method: types_1.ChromeCommand.RuntimeDisable,
        params: {},
    },
];
const log = new log_1.Logger('socket-server', enum_1.WinstonColor.Cyan);
/**
 * ws 调试服务，支持远程调试（无线模式，使用 ws 通道）、本地调试（有线模式，使用 tunnel 数据通道）
 * 通过 redis pub/sub 实现多点部署时的消息分发，上下行消息根据 clientId 来建立 redis channel 进行分发
 * 本地安装 npm 包部署时（单点部署），不存储 redis，而是直接保存在内存中
 *
 * 设计方案： https://iwiki.woa.com/pages/viewpage.action?pageId=1222336167
 */
class SocketServer {
    constructor(server) {
        this.server = server;
    }
    /**
     * 开启调试服务
     */
    start() {
        const wss = new ws_1.Server({
            server: this.server,
            path: config_1.config.wsPath,
        });
        this.wss = wss;
        wss.on('connection', this.onConnection.bind(this));
        wss.on('error', (e) => {
            var _a;
            log.error('wss error: %s', (_a = e) === null || _a === void 0 ? void 0 : _a.stack);
        });
        wss.on('headers', (headers) => {
            log.info('wss headers: %j', headers);
        });
        wss.on('upgrade', (response) => {
            log.info('wss upgrade: %j', response);
        });
    }
    /**
     * 关闭调试服务，并清理当前节点连接的调试对象缓存
     */
    close() {
        pub_sub_manager_1.cleanAllDebugTargets();
        this.wss.close(() => {
            log.info('wss closed.');
        });
    }
    onConnection(ws, req) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            log.info('on connection, ws url: %s', req.url);
            const wsUrlParams = url_1.parseWsUrl(req.url);
            const { clientRole, clientId } = wsUrlParams;
            if (clientRole === enum_1.ClientRole.Devtools) {
                this.onDevtoolsConnection(ws, wsUrlParams);
            }
            else {
                this.onAppConnection(ws, wsUrlParams);
            }
            /**
             * ⚠️ 给 ws 添加事件监听前，不要执行异步操作
             * 连接建立后，再判断是否合法，不合法则关闭
             */
            const debugTarget = yield debug_targets_1.DebugTargetManager.findDebugTarget(clientId);
            const isValid = url_1.isConnectionValid(wsUrlParams, debugTarget);
            if (!isValid)
                return ws.close();
        });
    }
    /**
     * 将来自于 devtools 的 ws，通过 pub/sub 转发到 redis
     */
    onDevtoolsConnection(ws, wsUrlParams) {
        const { Subscriber, Publisher } = db_1.getDBOperator();
        const { extensionName, clientId } = wsUrlParams;
        const downwardChannelId = pub_sub_channel_1.createDownwardChannel(clientId, extensionName);
        const upwardChannelId = pub_sub_channel_1.createUpwardChannel(clientId, extensionName);
        const internalChannelId = pub_sub_channel_1.createInternalChannel(clientId, '');
        log.info('devtools connected, subscribe channel: %s, publish channel: %s', downwardChannelId, upwardChannelId);
        const downwardSubscriber = new Subscriber(downwardChannelId);
        // internal channel 用于订阅 node 节点之间的事件，如 app ws close 时，通知 devtools ws close
        const internalSubscriber = new Subscriber(internalChannelId);
        const publisher = new Publisher(upwardChannelId);
        downwardSubscriber.unsubscribe();
        internalSubscriber.unsubscribe();
        downwardSubscriber.subscribe(ws.send.bind(ws));
        internalSubscriber.subscribe((msg) => {
            if (msg === enum_1.InternalChannelEvent.WSClose) {
                log.warn('close devtools ws connection');
                ws.close();
                internalSubscriber.unsubscribe();
                internalSubscriber.disconnect();
            }
        });
        ws.on('message', (msg) => {
            publisher.publish(msg.toString());
        });
        const onClose = (e) => {
            if (e)
                log.error('devtools ws client error %s', (e === null || e === void 0 ? void 0 : e.stack) || e);
            log.info('devtools ws disconnect. clientId: %s', clientId);
            resumeCommands.map(publisher.publish.bind(publisher));
            // 延时等 publisher 发布完成
            process.nextTick(() => {
                publisher.disconnect();
                downwardSubscriber.unsubscribe();
                downwardSubscriber.disconnect();
            });
        };
        ws.on('close', onClose);
        // ws 标准规定 on error 后一定触发 on close，所以也要做清理
        ws.on('error', onClose);
    }
    /**
     * app ws 连接，创建调试对象，并订阅上行调试消息
     */
    onAppConnection(ws, wsUrlParams) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { clientId, clientRole } = wsUrlParams;
            log.info('ws app client connected. %s', clientId);
            const platform = {
                [enum_1.ClientRole.Android]: "2" /* Android */,
                [enum_1.ClientRole.IOS]: "1" /* IOS */,
            }[clientRole];
            const useWS = client_1.appClientManager.shouldUseAppClientType(platform, "WsAppClient" /* WS */);
            if (!useWS)
                return log.warn('current env is %s, ignore ws connection', global.debugAppArgv.env);
            let debugTarget = debug_target_1.createTargetByWsUrlParams(wsUrlParams);
            // app ws 添加监听前可以执行异步操作，因为 app 建立连接后不会主动发送任何消息
            debugTarget = yield debug_target_1.patchDebugTarget(debugTarget);
            const { model } = db_1.getDBOperator();
            model.upsert(config_1.config.redis.key, clientId, debugTarget);
            pub_sub_manager_1.subscribeCommand(debugTarget, ws);
            const onClose = (e) => {
                if (e)
                    log.error('app ws error %s', e === null || e === void 0 ? void 0 : e.stack);
                log.info('ws app client disconnect. %s', clientId);
                pub_sub_manager_1.cleanDebugTarget(clientId);
            };
            ws.on('close', onClose);
            ws.on('error', onClose);
        });
    }
}
exports.SocketServer = SocketServer;
